import logging
import yaml
from os import environ


class Config(dict):
    """A class to manage the bot's configuration"""

    keys = ["BOT_SERVER", "BOT_USERNAME", "BOT_PASSWORD", "BOT_ACCESS_TOKEN",
            "API_BASE_URL", "API_TOKEN",
            "LOGGING_LEVEL"]

    def __init__(self, config_path=None):
        logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.DEBUG)

        """
        The config is generated via 4 different paths from lowest to highest priority
        4. a config.yml in the working directory of the bot
        3. a file in yml format specified vie --config
        2. a file in yml format specified via the CONFIG_PATH environment variable
        1. config values specified via their name as environment variable
        """
        # Try to find config via environment variable
        try:
            config_path = environ["CONFIG_PATH"]
        except KeyError:
            logging.debug(f"No config file set via the environment variable")
            pass
        if config_path is None:
            logging.debug(f"No config file set via the --config option, defaulting to config.yml in working directory")
            config_path = "config.yml"
        logging.info(f"Tying to load bot configuration from {config_path}")
        try:
            with open(config_path, 'r') as file:
                self.extend_by_dict(yaml.safe_load(file))
        except FileNotFoundError:
            logging.error(f"Cold not find bot configuration at {config_path}")


        """
        This maps all self.keys (e.g. "LOGGING_LEVEL") that are in the environment to the corresponding config value
        e.g. self["logging"]["level"]. Does not support more than 2 level
        """
        for key in self.keys:
            scope, k = [x.lower() for x in key.split("_", maxsplit=1)]
            try:
                environ[key]
            except KeyError:
                logging.debug(f"{key} not set in environment")
                continue
            try:
                self[scope]
            except KeyError:
                self[scope] = {}
            self[scope][k] = environ[key]
            logging.debug(f"{key} set via environment")

        """Set the logging level according to config"""
        if self["logging"]['level'] in ["debug", "DEBUG"]:
            logging_level = logging.DEBUG
        elif self["logging"]['level'] in ["error", "ERROR"]:
            logging_level = logging.ERROR
        else:
            logging_level = logging.INFO
        logging.getLogger().setLevel(logging_level)

        try:
            self["bot"]["prefix"]
        except KeyError:
            self["bot"]["prefix"] = ""

    def extend_by_dict(self, data):
        for key in data:
            self[key] = data[key]



import pytest
from matrix_registration_bot.registration_api import RegistrationAPI

valid_tokens = ["TrwUI5zHm~Gn3M9Am", "gpWrPaFrbuP73A6N", "dada", "a", "1", "J_2NGPksUSbST1cp",
                "uERKLWlzIDhrVxQCSGLSmBdMEnKnnaOCNBUawgLgUyjjqnaIBFmMkJQATTpqhbXX"]
invalid_tokens = ["dajaj/aeofjj", "<script>alert(0)</script>", "&lt;script&gt;alert(&#39;1&#39;);&lt;/script&gt;",
                  "ɐuƃɐɯ", "register!", "uERKLWlzIDhrVxQCSGLSmBdMEnKnnaOCNBUawgLgUyjjqnaIBFmMkJQXAT65chars"]


def test_valid_token_format():
    for token in valid_tokens:
        if not RegistrationAPI.valid_token_format(token):
            raise AssertionError(f"Falsely said {token} is a invalid token")
    for token in invalid_tokens:
        if RegistrationAPI.valid_token_format(token):
            raise AssertionError(f"Falsely said {token} is a valid token")



import re
from datetime import datetime, timedelta
import aiohttp


class RegistrationAPI:
    def __init__(self, base_url: str, endpoint: str, api_token: str):
        self.base_url = base_url
        self.endpoint = endpoint
        self.api_token = api_token
        self.headers = {"Authorization": f"Bearer {api_token}"}
        self.session = None

    def __str__(self):
        return f"API Connection to {self.base_url}"

    async def ensure_session(self):
        if self.session is None:
            self.session = aiohttp.ClientSession(self.base_url)

    @staticmethod
    def verbose_response(r):
        return f"The registration api returned `{r.status}: {r.reason}` for {r.method}: {r.url}"

    @staticmethod
    def check_response(r):
        if r.status == 404:
            raise FileNotFoundError("Token not found or API not reachable (404 Not Found)")
        elif r.status == 401:
            raise PermissionError(RegistrationAPI.verbose_response(r) +
                                  f" Check, that the API access token is correct")
        elif r.status != 200:
            raise ConnectionError(RegistrationAPI.verbose_response(r))



    @staticmethod
    def token_to_markdown(token_details: dict):
        """
        Converts a token to markdown string

        :param token_details: A dictionary containing the token
            Example: {'token': '8iB~zWiDU1SC0NT3',
                      'uses_allowed': 1,
                      'pending': 0,
                      'completed': 1,
                      'expiry_time': 1642807497388}
        :return: a string in markdown format
        """
        if token_details['uses_allowed'] is None:
            uses_left = "Unlimited"
        else:
            uses_left = token_details['uses_allowed'] - (token_details['completed'] + token_details['pending'])
        if token_details['expiry_time'] is None:
            timestamp = "Does not expire"
        else:
            datetime_obj = datetime.utcfromtimestamp(int(token_details['expiry_time']) / 1000)
            timestamp = datetime_obj.strftime("%d.%m.%y %H:%M UTC")
        md = (f"""**Token:** `{token_details['token']}`
        Expires: {timestamp}
        Uses left: {uses_left}
        """)
        return md

    @staticmethod
    def token_to_short_markdown(token_details: dict):
        """
        Converts a token to markdown string

        :param token_details: A dictionary containing the token
            Example: {'token': '8iB~zWiDU1SC0NT3',
                      'uses_allowed': 1,
                      'pending': 0,
                      'completed': 1,
                      'expiry_time': 1642807497388}
        :return: a string of only the token value in markdown format
        """
        md = f"`{token_details['token']}`"
        return md

    @staticmethod
    def valid_token_format(token: str):
        """
        Checks if a string is a valid token.

        The string is checked against a regex pattern. Due to the restricted nature of the token format, the string is
        safe to use when in this format. More information: https://matrix-org.github.io/synapse/latest/usage/administration/admin_api/registration_tokens.html#create-token
        :param token: The token value to check.
        :return:bool: True if the token is in valid format, else false
        """
        if len(token) > 64:
            return False
        pattern = re.compile("[A-Za-z0-9._~-]*")
        match = re.fullmatch(pattern, token)
        return match

    async def list_tokens(self):
        """
        Gathers a list of all registration tokens

        :return: List of token_details
        """
        await self.ensure_session()
        async with self.session.get(self.endpoint, headers=self.headers) as r:
            try:
                assert r.status == 200
            except AssertionError:
                raise ConnectionError(self.verbose_response(r))
            return (await r.json())["registration_tokens"]

    async def get_token(self, token):
        """
        Gets token

        :return: token_details as dict
        """
        await self.ensure_session()
        if self.valid_token_format(token):
            async with self.session.get(self.endpoint + f"/{token}", headers=self.headers) as r:
                self.check_response(r)
                return await r.json()
        else:
            raise TypeError("Token is not a valid format!")

    async def delete_all_token(self):
        """
        Deletes all token

        :return: List of deleted token
        """
        await self.ensure_session()
        all_tokens = await self.list_tokens()
        for token in all_tokens:
            await self.delete_token(token["token"])
        return all_tokens

    async def delete_token(self, token: str):
        """
        Deletes the given token

        :param token:
        :return: The token_details that is deleted as dict
        """
        await self.ensure_session()
        if self.valid_token_format(token):
            r = await self.session.get(f"{self.endpoint}/{token}", headers=self.headers)
            self.check_response(r)
            token_details = await r.json()
            async with self.session.delete(f"{self.endpoint}/{token}", headers=self.headers) as r:
                self.check_response(r)
                return token_details
        else:
            raise ValueError(f"Token {token} is not a valid format!")

    async def create_token(self, expiry_days=7):
        """
        Create a token for registering a user

        expire_days:int
            Determines how long the token will be valid (in days)
        :return: token_details
        """
        await self.ensure_session()
        expiry_time = int(datetime.timestamp(datetime.now() + timedelta(days=expiry_days)) * 1000)
        data = '{"uses_allowed": 1, "expiry_time": ' + str(expiry_time) + '}'
        async with self.session.post(f"{self.endpoint}/new", data=data, headers=self.headers) as r:
            self.check_response(r)
            return await r.json()



import cryptography
import simplematrixbotlib as botlib
import matrix_registration_bot
from matrix_registration_bot.registration_api import RegistrationAPI
from matrix_registration_bot.config import Config
import logging
import argparse

parser = argparse.ArgumentParser(description='Start the matrix-registration-bot.')

parser.add_argument('--config', default=None, help='Specify a configuration file to use')

args = parser.parse_args()
if args.config is None:
    config_path = 'config.yml'
else:
    config_path = args.config
config = Config(args.config)

bot_server = config['bot']['server']
bot_username = config['bot']['username']
try:
    bot_access_token = config['bot']['access_token']
    creds = botlib.Creds(bot_server,
                         username=bot_username,
                         access_token=bot_access_token)
except KeyError:
    logging.info("Using password based authentication for the bot")
    try:
        bot_access_password = config['bot']['password']
    except KeyError:
        error = "No access token or password for the bot provided"
        logging.error(error)
        raise KeyError(error)
    creds = botlib.Creds(bot_server,
                         username=bot_username,
                         password=bot_access_password)

bot_prefix = config['bot']['prefix']

api_base_url = config['api']['base_url']
api_endpoint = '/_synapse/admin/v1/registration_tokens'
api_token = config['api']['token']

# Load a config file that configures bot behaviour
smbl_config = botlib.Config()
smbl_config.emoji_verify = True
smbl_config.ignore_unverified_devices = True
SIMPLE_MATRIX_BOT_CONFIG_FILE = "config.toml"
try:
    smbl_config.load_toml(SIMPLE_MATRIX_BOT_CONFIG_FILE)
    logging.info(f"Loaded the simple-matrix-bot config file {SIMPLE_MATRIX_BOT_CONFIG_FILE}")
except FileNotFoundError:
    logging.info(f"No simple-matrix-bot config file found. Creating {SIMPLE_MATRIX_BOT_CONFIG_FILE}")
    smbl_config.save_toml(SIMPLE_MATRIX_BOT_CONFIG_FILE)

bot = botlib.Bot(creds, smbl_config)

api = RegistrationAPI(api_base_url, api_endpoint, api_token)

help_string = (
    f"""**[Matrix Registration Bot](https://github.com/moan0s/matrix-registration-bot/)** {matrix_registration_bot.__version__}
You can always ask for help in
[#matrix-registration-bot:hyteck.de](https://matrix.to/#/#matrix-registration-bot:hyteck.de)!

**Unrestricted commands**

* `{bot_prefix}help`: Shows this help

**Restricted commands**

* `{bot_prefix}list`: Lists all registration tokens
* `{bot_prefix}show <token>`: Shows token details in human-readable format
* `{bot_prefix}create`: Creates a token that that is valid for one registration for seven days
* `{bot_prefix}delete <token>` Deletes the specified token(s)
* `{bot_prefix}delete-all` Deletes all tokens
* `{bot_prefix}allow @user:example.com` Allows the specified user (or a user matching a regex pattern) to use restricted commands
* `{bot_prefix}disallow @user:example.com` Stops a specified user (or a user matching a regex pattern) from using restricted commands
""")


@bot.listener.on_message_event
async def token_actions(room, message):
    match = botlib.MessageMatch(room, message, bot, bot_prefix)
    # Unrestricted commands
    if match.is_not_from_this_bot() and match.contains("help") and match.prefix():
        """The help command should be accessible even to users that are not allowed"""
        logging.info(f"{match.event.sender} viewed the help")
        await bot.api.send_markdown_message(room.room_id, help_string)
    # Restricted commands
    elif match.is_not_from_this_bot() and match.is_from_allowed_user() and match.prefix():
        if match.command("list"):
            logging.info(f"{match.event.sender} listed all tokens")
            try:
                token_list = await api.list_tokens()
            except ConnectionError as e:
                logging.warning(f"Error while trying to list all tokens: {e}")
                await error_handler(room, e)
                return
            if len(token_list) < 10:
                message = "\n".join([RegistrationAPI.token_to_markdown(token) for token in token_list])
            else:
                tokens_as_string = [RegistrationAPI.token_to_short_markdown(token) for token in token_list]
                message = f"All tokens: {', '.join(tokens_as_string)}"
            await bot.api.send_markdown_message(room.room_id, message)

        if match.command("create"):
            try:
                token = await api.create_token()
                logging.info(f"{match.event.sender} created token {token}")
                await bot.api.send_markdown_message(room.room_id, f"{RegistrationAPI.token_to_markdown(token)}")
            except (ConnectionError, PermissionError, FileNotFoundError) as e:
                logging.warning(f"Error while trying to create a token: {e}")
                await error_handler(room, e)

        if match.command("delete-all"):
            deleted_tokens = await api.delete_all_token()
            logging.info(f"{match.event.sender} deleted all tokens")
            await send_info_on_deleted_token(room, deleted_tokens)

        if match.command("delete"):
            deleted_tokens = []
            logging.info(f"{match.event.sender} tries to delete {match.args()}")
            if not len(match.args()) > 0:
                await bot.api.send_markdown_message(room.room_id, "You must give a token!")
            for token in match.args():
                token = token.strip()
                try:
                    deleted_tokens.append(await api.delete_token(token))
                except ValueError as e:
                    logging.info(f"Token {token} given by {match.event.sender} to delete was not in correct format")
                    await error_handler(room, e)
                except FileNotFoundError as e:
                    logging.info(f"Token {token} given by {match.event.sender} to delete was not found")
                    await error_handler(room, e)
                except ConnectionError as e:
                    logging.warning(f"Error: {e} while trying to get a token")
                    await error_handler(room, e)
            logging.info(f"{match.event.sender} deleted token {deleted_tokens}")
            await send_info_on_deleted_token(room, deleted_tokens)

        if match.command("show"):
            tokens_info = []
            logging.info(f"{match.event.sender} tries to show {match.args()}")
            if not len(match.args()) > 0:
                await bot.api.send_markdown_message(room.room_id, "You must give a token!")
                return
            for token in match.args():
                token = token.strip()
                try:
                    token_info = await api.get_token(token)
                    logging.info(f"Showing {token} to {match.event.sender}")
                    tokens_info.append(RegistrationAPI.token_to_markdown(token_info))
                except ConnectionError as e:
                    logging.warning(f"Error while trying to get a token: {e}")
                    await error_handler(room, e)
                except FileNotFoundError as e:
                    logging.info(f"Token {token} given by {match.event.sender} to show was not found")
                    await error_handler(room, e)
                except TypeError as e:
                    logging.info(f"Token {token} given by {match.event.sender} to show was not in correct format")
                    await error_handler(room, e)
            if len(tokens_info) > 0:
                await bot.api.send_markdown_message(room.room_id, "\n".join(tokens_info))

        if match.command("allow"):
            bot.config.add_allowlist(set(match.args()))
            bot.config.save_toml("config.toml")
            logging.info(f"{match.event.sender} allowed {set(match.args())} (if valid)")
            await bot.api.send_text_message(
                room.room_id,
                f'allowing {", ".join(arg for arg in match.args())} (if valid)')

        if match.command("disallow"):
            bot.config.remove_allowlist(set(match.args()))
            bot.config.save_toml("config.toml")
            logging.info(f"{match.event.sender} disallowed {set(match.args())} (if valid)")
            await bot.api.send_text_message(
                room.room_id,
                f'disallowing {", ".join(arg for arg in match.args())} (if valid)')


async def send_info_on_deleted_token(room, token_list):
    if len(token_list) > 0:
        message = f"Deleted the following token(s): "
        tokens_as_string = [RegistrationAPI.token_to_short_markdown(token) for token in token_list]
        message += ", ".join(tokens_as_string)
    else:
        message = "No token deleted"
    await bot.api.send_markdown_message(room.room_id, message)


async def error_handler(room, error):
    message = f"The bot encountered the following error:\n"
    message += error.args[0]
    await bot.api.send_markdown_message(room.room_id, message)


def run_bot():
    try:
        bot.run()
    except cryptography.fernet.InvalidToken:
        logging.error("The token does not seem to fit the saved session. this can happen if you change the bot user."
                      "If this is the case, deleting the session.txt and restarting the bot might help")
        exit(1)


if __name__ == "__main__":
    run_bot()
